name = 'async_sched'
version = '1.0.5'
description = 'Use async to have functions run on a schedule.'
url = 'https://github.com/justengel/async_sched'
author = 'Justin Engel'
author_email = 'jtengel08@gmail.com'
